/*
 * Copyright 2011, Maarten Baert <maarten-baert@hotmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#include "MemoryBuffer.h"

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <new>

// This is only possible because the interface is in the same
// DLL, if this was a separate DLL you would use a GM function
// to get the interface pointer.
extern SharedStreamBufferInterface ssbm_interface;

SharedBufferInterface sbi_memorybuffer;

// This class initializes the buffer interface
class MemoryBufferController {
	
	public:
	MemoryBufferController() {
		sbi_memorybuffer.readdata = &MemoryBuffer::ReadData;
		sbi_memorybuffer.writedata = &MemoryBuffer::WriteData;
		sbi_memorybuffer.getbytesleft = &MemoryBuffer::GetBytesLeft;
		sbi_memorybuffer.getreadpos = &MemoryBuffer::GetReadPos;
		sbi_memorybuffer.getwritepos = &MemoryBuffer::GetWritePos;
		sbi_memorybuffer.getlength = &MemoryBuffer::GetLength;
		sbi_memorybuffer.setreadpos = &MemoryBuffer::SetReadPos;
		sbi_memorybuffer.setwritepos = &MemoryBuffer::SetWritePos;
		sbi_memorybuffer.setlength = &MemoryBuffer::SetLength;
	}
	
} memorybuffercontroller;

void MemoryBuffer::Clear() {
	data = NULL;
	length = 0;
	capacity = 0;
	readpos = 0;
	writepos = 0;
}

MemoryBuffer* MemoryBuffer::Create() {
	MemoryBuffer *buffer = new MemoryBuffer();
	buffer->interface = &sbi_memorybuffer;
	ssbm_interface.RegisterBuffer(buffer);
	buffer->Clear();
	return buffer;
}

void MemoryBuffer::Destroy(MemoryBuffer* buffer) {
	ssbm_interface.UnregisterBuffer(buffer);
	free(buffer->data);
	delete buffer;
}

bool MemoryBuffer::ReadFromFile(const char* filename) {
	Clear();
	FILE *f = fopen(filename, "rb");
	if(f == NULL) return false;
	fseek(f, 0, SEEK_END);
	long size = ftell(f);
	fseek(f, 0, SEEK_SET);
	if(size < 0l) {
		fclose(f);
		return false;
	}
	try {
		SetLength(this, size);
	} catch(...) {
		fclose(f);
		throw;
	}
	if(fread(data, 1, size, f) != (size_t)(size)) {
		Clear();
		fclose(f);
		return false;
	}
	fclose(f);
	return true;
}

bool MemoryBuffer::WriteToFile(const char* filename) {
	FILE *f = fopen(filename, "wb");
	if(f == NULL) return false;
	if(fwrite(data, 1, length, f) != length) {
		fclose(f);
		return false;
	}
	fclose(f);
	return true;
}

void MemoryBuffer::ReadData(SharedStream* stream, void* ptr, uint32_t bytes) {
	MemoryBuffer *b = static_cast<MemoryBuffer*>(stream);
	if(bytes > b->length - b->readpos) {
		memset(ptr, 0, bytes);
	} else {
		memcpy(ptr, b->data + b->readpos, bytes);
		b->readpos += bytes;
	}
}

void MemoryBuffer::WriteData(SharedStream* stream, const void* ptr, uint32_t bytes) {
	MemoryBuffer *b = static_cast<MemoryBuffer*>(stream);
	if(bytes > b->length - b->writepos) {
		SetLength(b, b->writepos + bytes);
	}
	memcpy(b->data + b->writepos, ptr, bytes);
	b->writepos += bytes;
}

uint32_t MemoryBuffer::GetBytesLeft(const SharedStream* stream) {
	const MemoryBuffer *b = static_cast<const MemoryBuffer*>(stream);
	return b->length - b->readpos;
}

uint32_t MemoryBuffer::GetReadPos(const SharedBuffer* buffer) {
	const MemoryBuffer *b = static_cast<const MemoryBuffer*>(buffer);
	return b->readpos;
}

uint32_t MemoryBuffer::GetWritePos(const SharedBuffer* buffer) {
	const MemoryBuffer *b = static_cast<const MemoryBuffer*>(buffer);
	return b->writepos;
}

uint32_t MemoryBuffer::GetLength(const SharedBuffer* buffer) {
	const MemoryBuffer *b = static_cast<const MemoryBuffer*>(buffer);
	return b->length;
}

void MemoryBuffer::SetReadPos(SharedBuffer* buffer, uint32_t pos) {
	MemoryBuffer *b = static_cast<MemoryBuffer*>(buffer);
	if(pos > b->length) {
		b->readpos = b->length;
	} else {
		b->readpos = pos;
	}
}

void MemoryBuffer::SetWritePos(SharedBuffer* buffer, uint32_t pos) {
	MemoryBuffer *b = static_cast<MemoryBuffer*>(buffer);
	if(pos > b->length) {
		b->writepos = b->length;
	} else {
		b->writepos = pos;
	}
}

void MemoryBuffer::SetLength(SharedBuffer* buffer, uint32_t length) {
	MemoryBuffer *b = static_cast<MemoryBuffer*>(buffer);
	// the minimum size is 16
	// this will result in a capacity between 20 and 31 (inclusive)
	unsigned int required = length;
	if(required < 16) required = 16;
	unsigned int new_capacity;
	if(required > b->capacity) {
		// allocate 50% more than required
		new_capacity = required + required / 2;
	} else if(required <= b->capacity / 2) {
		// keep 25% more than required
		new_capacity = required + required / 4;
	} else {
		// no reallocation
		b->length = length;
		if(b->readpos > length) {
			b->readpos = length;
		}
		if(b->writepos > length) {
			b->writepos = length;
		}
		return;
	}
	void *temp = realloc(b->data, new_capacity);
	if(temp==NULL) {
		throw std::bad_alloc();
	}
	b->data = (char*)(temp);
	if(length > b->length) {
		memset(b->data + b->length, 0, length - b->length);
	}
	b->length = length;
	b->capacity = new_capacity;
	if(b->readpos > length) {
		b->readpos = length;
	}
	if(b->writepos > length) {
		b->writepos = length;
	}
}

