/*
 * Copyright 2011, Maarten Baert <maarten-baert@hotmail.com>
 * 
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef SHAREDSTREAMBUFFERS_H
#define SHAREDSTREAMBUFFERS_H

#include <stdint.h>

class SharedStream;
class SharedBuffer;

class SharedStreamBufferInterface {
	
	friend class SharedStreamBufferManager;
	
	private:
	void (*registerstream)(SharedStream* stream);
	void (*unregisterstream)(SharedStream* stream);
	SharedStream* (*findstream)(uint32_t id);
	void (*registerbuffer)(SharedBuffer* buffer);
	void (*unregisterbuffer)(SharedBuffer* buffer);
	SharedBuffer* (*findbuffer)(uint32_t id);
	
	public:
	inline void RegisterStream(SharedStream* stream) { (*registerstream)(stream); }
	inline void UnregisterStream(SharedStream* stream) { (*unregisterstream)(stream); }
	inline SharedStream* FindStream(uint32_t id) { return (*findstream)(id); }
	inline void RegisterBuffer(SharedBuffer* buffer) { (*registerbuffer)(buffer); }
	inline void UnregisterBuffer(SharedBuffer* buffer) { (*unregisterbuffer)(buffer); }
	inline SharedBuffer* FindBuffer(uint32_t id) { return (*findbuffer)(id); }
	
};

struct SharedStreamInterface {
	void (*readdata)(SharedStream*, void*, uint32_t);
	void (*writedata)(SharedStream*, const void*, uint32_t);
	uint32_t (*getbytesleft)(const SharedStream*);
};

class SharedStream {
	
	friend class SharedStreamBufferManager;
	
	private:
	uint32_t id;
	
	protected:
	SharedStreamInterface *interface;
	
	public:
	inline uint32_t GetID() { return id; }
	inline SharedStreamInterface* GetInterface() { return interface; } // can be used to check the actual type of the buffer/stream
	inline void ReadData(void* ptr, uint32_t bytes) { (*interface->readdata)(this, ptr, bytes); }
	inline void WriteData(const void* ptr, uint32_t bytes) { (*interface->writedata)(this, ptr, bytes); }
	inline uint32_t GetBytesLeft() const { return (*interface->getbytesleft)(this); }
	
};

struct SharedBufferInterface : public SharedStreamInterface {
	uint32_t (*getreadpos)(const SharedBuffer*);
	uint32_t (*getwritepos)(const SharedBuffer*);
	uint32_t (*getlength)(const SharedBuffer*);
	void (*setreadpos)(SharedBuffer*, uint32_t);
	void (*setwritepos)(SharedBuffer*, uint32_t);
	void (*setlength)(SharedBuffer*, uint32_t);
};

class SharedBuffer : public SharedStream {
	
	friend class SharedStreamBufferManager;
	
	public:
	inline uint32_t GetReadPos() const { return (*static_cast<SharedBufferInterface*>(interface)->getreadpos)(this); }
	inline uint32_t GetWritePos() const { return (*static_cast<SharedBufferInterface*>(interface)->getwritepos)(this); }
	inline uint32_t GetLength() const { return (*static_cast<SharedBufferInterface*>(interface)->getlength)(this); }
	inline void SetReadPos(uint32_t pos) { (*static_cast<SharedBufferInterface*>(interface)->setreadpos)(this, pos); }
	inline void SetWritePos(uint32_t pos) { (*static_cast<SharedBufferInterface*>(interface)->setwritepos)(this, pos); }
	inline void SetLength(uint32_t length) { (*static_cast<SharedBufferInterface*>(interface)->setlength)(this, length); }
	
};

#endif // SHAREDSTREAMBUFFERS_H

